/* ============================================================================
 * TTCPServer.java
 * ============================================================================
 * THIS SOFTWARE IS RELEASED UNDER THE FOLLOWING BINDING LEGAL AGREEMENT
 * (BSD-LICENSE).
 * ============================================================================
 *
 * Copyright (C) 2007-2009
 * by J.H.H.C. van der Meijs (tcpconns@jhhcvandermeijs.nl)
 * 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 *    * Redistributions of source code must retain the above copyright notice, 
 *      this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright 
 *      notice, this list of conditions and the following disclaimer in the 
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the name of tcpconns nor the names of its copyright holders or 
 *      its contributors may be used to endorse or promote products derived from 
 *      this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *
 * ============================================================================
 * IF YOU DO NOT AGREE TO THE ABOVE TERMS AND CONDITIONS THEN DO *NOT* USE THIS 
 * SOFTWARE AND *DO* DESTROY ALL COPIES OF THIS SOFTWARE THAT YOU HAVE.
 * ============================================================================
 * 
 * TCPCONNS’ CURRENT CODE IS NOT KNOWN TO VIOLATE ANY COPYRIGHTS OR INFRINGE ON 
 * ANY PATENTS. MOREOVER, IT ADHERES TO GENERAL PROGRAMMING RULES AND PRACTICES 
 * AND IS OF SUCH A SIMPLISTIC NATURE THAT A SIMILAR PACKAGE COULD BE 
 * CONSTRUCTED BY ANYONE. IMPORTANT(!): IF YOU FEEL THAT A PIECE OF CODE IN
 * TCPCONNS VIOLATES A COPYRIGHT OR INFRINGES ON A PATENT, PLEASE DO INFORM ONE 
 * OF THE AUTHORS IMMEDIATELY. DO ALWAYS STATE EXACTLY WHAT CODE IS NOT ALLOWED
 * UNDER EXACTLY WHAT RULING, COPYRIGHT, OR PATENT; AND PLEASE DO ELABORATE ON 
 * EXACTLY HOW THE CODE VIOLATES THE SPECIFIC COPYRIGHT OR INFRINGES ON
 * THE SPECIFIC PATENT.
 * 
 * ============================================================================
 * 
 * This software is Open Source Initiative approved Open Source Software.
 * Open Source Initiative Approved is a trademark of the Open Source Initiative.
 * http://www.opensource.org/
 * 
 * ============================================================================
 */
package nl.jhhcvandermeijs.tcpconns.server;

import nl.jhhcvandermeijs.tcpconns.client.*;
import java.io.IOException;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

/**
 * this class implements a multi-threaded tcp server
 */
public class TTCPServer {      
       
    /**
     * placeholder for the socket timeout value
     */
    protected int fTimeOut = 10*1000;
    
    /**
     * use this method to set the socket timeout value
     * @param timeOut time in milliseconds; default == 10*1000 ms
     */
    synchronized public void setTimeOut(int timeOut) {
        if (timeOut <= 0) {timeOut = 10*1000;}
        this.fTimeOut = timeOut;
    }
    
    /**
     * use this method to read the socket timeout value
     * @return timeOut time in milliseconds; default == 10*1000 ms
     */
    synchronized public int getTimeOut() {
        return this.fTimeOut;
    }
    
    /**
     * place holder for the maximum number of clients that is allowed to
     * connect to this server simultaneously
     */
    protected int fMaxClients = 256;
    
    /**
     * use this method to set the maximum number of clients that is allowed
     * to connect to this server simultaneously
     * @param maxClients the number of clients that is permitted to be connected
     * to this server concurrently
     */
    synchronized public void setMaxClients(int maxClients) {
        if (maxClients <= 0) {maxClients = 256;}
        this.fMaxClients = maxClients;
    }
    /**
     * use this method to read the maximum number of clients that is permitted
     * to be connected to this server all at once
     * @return the maximum number of clients that is allowed to be connected
     * to this server simultaneously
     */
    synchronized public int getMaxClients() {
        return this.fMaxClients;
    }
    
    /**
     * override this method so that it returns instances of your TServerSideProtocol
     * class
     * 
     * @param aTTCPConnection a parameter that contains the link to a
     * connected client
     * @param aTTCPServer the parameter that contains the link to your
     * TTCPServer
     * @return an object that is of your TServerSideProtocol class
     * @throws java.io.IOException
     */
    synchronized public TServerSideProtocol newServerSideProtocol(
            TTCPConnection aTTCPConnection, TTCPServer aTTCPServer) 
            throws IOException {
            throw new IOException();
    }
    
    /**
     * counter for the number of clients that is currently connected
     */
    protected int fClientCount = 0;
    public synchronized int getClientCount() {
        return this.fClientCount;
    }
    /**
     * synchronized method that increases the number of actively connected
     * clients
     */    
    synchronized void incClientCount() {
        this.fClientCount++;
    }
    /**
     * synchronized method to decrease the number of concurrently connected
     * clients
     */
    synchronized void decClientCount() {
        this.fClientCount--;
    }
    
    /**
     * reference to the ServerSocket
     */
    protected ServerSocket fmyServerSocket = null;
    
    /**
     * handle towards the threadpool, which is a cached thread pool instance
     */
    protected ExecutorService fmyThreadPool = null;
    
    /**
     * this method is used to handoff the connected client to the threadpool
     * 
     * @param aTTCPConnection link towards an active connection
     */
    synchronized void serviceClient(TTCPConnection aTTCPConnection) {
        if (this.fClientCount >= this.fMaxClients) {
            try {aTTCPConnection.mySocket().close();} catch (Exception e) {}
        } else { //if then else
            this.incClientCount();
            try {
                    TServerSideProtocol aClient = 
                        this.newServerSideProtocol(aTTCPConnection, this);            
                    this.fmyThreadPool.execute(aClient);
            } catch (Exception e) {
                try {
                    aTTCPConnection.mySocket().close();
                } catch (Exception e2) {
                } finally {
                    this.decClientCount();
                }
            } 
        } //if then else
    }
    
    /**
     * holds the code that listens for new connections; is not actually a
     * thread but a Runnable implementation
     */
    protected class ListenerThread implements Runnable {
        protected TTCPServer fmyTTCPServer = null;
        public ListenerThread(TTCPServer myTTCPServer) {
            this.fmyTTCPServer = myTTCPServer;
        }
        public void run() {
            if (this.fmyTTCPServer == null) {return;}
            ServerSocket myServerSocket = this.fmyTTCPServer.fmyServerSocket;            
            try {
                myServerSocket.setSoTimeout(this.fmyTTCPServer.getTimeOut());
            } catch (SocketException ex) {return;} 
            Socket incomingClient = null;
            TTCPConnection client = null;
            while (!(Thread.currentThread().isInterrupted())
                   &&
                   (this.fmyTTCPServer.isActive())
                    ) {
                try {
                    incomingClient = myServerSocket.accept();
                    client = new TTCPConnection(false);
                    client.connect(incomingClient);
                    this.fmyTTCPServer.serviceClient(client);
                } catch (Exception e) {}
            }
            try {
                myServerSocket.close();
            } catch (Exception e) {}
        }        
    }
    /**
     * handle towards the ListenerThread
     */
    protected ListenerThread fmyListenerThread = null;

    /**
     * field to hold the server's IP binding
     */
    private String fBindingIP = null;
    /**
     * method to get fIP
     * @return fIP
     */
    public synchronized String getBindingIP() {return this.fBindingIP;}
    public synchronized void setBindingIP(String bindingIP) {this.fBindingIP = bindingIP;}
    /**
     * field to hold the server's port binding
     */
    private int fBindingPort = -1;
    /**
     * method to obtain fPort
     * @return fPort
     */
    public synchronized int getBindingPort() {return this.fBindingPort;}
    public synchronized void setBindingPort(int bindingPort) {this.fBindingPort = bindingPort;}

    public void activate() throws IOException {        
        this.activate(this.getBindingIP(), this.getBindingPort());        
    }

    /**
     * use this method to start the server
     * @param ip e.g. "0.0.0.0" to start serving on all local (ipv4) addresses
     * @param port the port on which the server should serve/listen for clients
     * @throws java.io.IOException
     */
    public void activate(String ip, int port) 
        throws IOException {
        this.deactivate();
        this.fBindingIP = ip;
        this.fBindingPort = port;
        this.fClientCount = 0;
        this.fmyServerSocket = 
            new ServerSocket(port, this.fMaxClients, InetAddress.getByName(ip));
        this.fmyThreadPool = Executors.newCachedThreadPool();
        this.fmyListenerThread = new ListenerThread(this);
        this.fmyThreadPool.execute(this.fmyListenerThread);
    }
    
    /**
     * use this method to stop serving new clients and to stop serving active
     * clients as soon as possible
     */
    public void deactivate() {
        if (this.fmyThreadPool != null) {
            this.fmyThreadPool.shutdown();
        }
        //this.fmyThreadPool = null;
        this.fmyListenerThread = null;
        if (this.fmyServerSocket != null) {
            try{this.fmyServerSocket.close();} catch (Exception e) {}
        }
        this.fmyServerSocket = null;
    }
    
    /**
     * try to force an immediate withdrawal from all active serving
     */
    public void deactivateNow() {
        if (this.fmyThreadPool != null) {
            this.fmyThreadPool.shutdownNow();
        }
        this.fmyThreadPool = null;
        this.fmyListenerThread = null;
        if (this.fmyServerSocket != null) {
            try{this.fmyServerSocket.close();} catch (Exception e) {}
        }
        this.fmyServerSocket = null;
    }
    
    /**
     * tests whether the server is active; Important(!): assumes no problems
     * occur in the ListenerThread
     * @return true if actively serving
     */
    synchronized public boolean isActive() {
        if (this.fmyListenerThread != null) {
            return true;
        } else {
            return false;
        }
    }
}
