/* ============================================================================
 * TTLSProvider.java
 * ============================================================================
 * THIS SOFTWARE IS RELEASED UNDER THE FOLLOWING BINDING LEGAL AGREEMENT
 * (BSD-LICENSE).
 * ============================================================================
 *
 * Copyright (C) 2007-2009
 * by J.H.H.C. van der Meijs (tcpconns@jhhcvandermeijs.nl)
 * 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 *    * Redistributions of source code must retain the above copyright notice, 
 *      this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright 
 *      notice, this list of conditions and the following disclaimer in the 
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the name of tcpconns nor the names of its copyright holders or 
 *      its contributors may be used to endorse or promote products derived from 
 *      this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED 
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *
 * ============================================================================
 * IF YOU DO NOT AGREE TO THE ABOVE TERMS AND CONDITIONS THEN DO *NOT* USE THIS 
 * SOFTWARE AND *DO* DESTROY ALL COPIES OF THIS SOFTWARE THAT YOU HAVE.
 * ============================================================================
 * 
 * TCPCONNS’ CURRENT CODE IS NOT KNOWN TO VIOLATE ANY COPYRIGHTS OR INFRINGE ON 
 * ANY PATENTS. MOREOVER, IT ADHERES TO GENERAL PROGRAMMING RULES AND PRACTICES 
 * AND IS OF SUCH A SIMPLISTIC NATURE THAT A SIMILAR PACKAGE COULD BE 
 * CONSTRUCTED BY ANYONE. IMPORTANT(!): IF YOU FEEL THAT A PIECE OF CODE IN
 * TCPCONNS VIOLATES A COPYRIGHT OR INFRINGES ON A PATENT, PLEASE DO INFORM ONE 
 * OF THE AUTHORS IMMEDIATELY. DO ALWAYS STATE EXACTLY WHAT CODE IS NOT ALLOWED
 * UNDER EXACTLY WHAT RULING, COPYRIGHT, OR PATENT; AND PLEASE DO ELABORATE ON 
 * EXACTLY HOW THE CODE VIOLATES THE SPECIFIC COPYRIGHT OR INFRINGES ON
 * THE SPECIFIC PATENT.
 * 
 * ============================================================================
 * 
 * This software is Open Source Initiative approved Open Source Software.
 * Open Source Initiative Approved is a trademark of the Open Source Initiative.
 * http://www.opensource.org/
 * 
 * ============================================================================
 */
package nl.jhhcvandermeijs.tcpconns.tls;

import nl.jhhcvandermeijs.tcpconns.server.TServerSideProtocol;
import nl.jhhcvandermeijs.tcpconns.client.*;
import java.io.IOException;
import java.net.Socket;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.security.cert.X509Certificate;
import javax.net.ssl.KeyManager;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSocket;
import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.TrustManager;

/**
 * instances of this class are used to establish SSL connections layered over
 * already connected TTCPConnection items; it is not necessary to use this
 * class directly: utilize the methods of TTCPConnection instead
 */
public class TTLSProvider {
        /**
         * this trustmanager always accepts all certificates without question;
         * Important (!): do *NOT* use unless you know what you are doing - I am 
         * NOT kidding!!! You should verify certs as appropriate for your 
         * protocol and intention; and do verify certificate revocation status
         * as well!
         */
        static public class TDummyCATrustManager implements javax.net.ssl.X509TrustManager {    
            /** 
             * @return empty array
             */
            public java.security.cert.X509Certificate[] getAcceptedIssuers() {                        
                return new X509Certificate[0];
            }
            /** 
             * cert is valid always
             */
            public void checkClientTrusted(java.security.cert.X509Certificate[] chain, String authType) throws java.security.cert.CertificateException {
                //throw no exception
            } //checkClientTrusted
            /** 
             * cert is valid always
             */        
            public void checkServerTrusted(java.security.cert.X509Certificate[] chain, String authType) throws java.security.cert.CertificateException {
                //throw new exception
            }       
        } //public class TDummyCATrustManager

        /**
         * handle towards the TTCPConnection object associated with this
         * TTLSProvider object
         */
        protected TTCPConnection fmyTTCPConnection = null;
        
        /**
         * boolean to indicate the TTLSProvider should act as a client 
         * (fClientSide == true) or as a server (fClientSide == false)
         */
        protected boolean fClientSide = false;
        
        /**
         * constructor
         * @param aTTCPConnection the TTCPConnection object that is to be
         *        associated with this TTLSProvider object
         * @param clientSide should be true if TTCPConnection is a client
         *        connection and it should be false if this TTLSProvider
         *        is used at the server's end
         */
        public TTLSProvider(TTCPConnection aTTCPConnection, boolean clientSide) {
            this.fmyTTCPConnection = aTTCPConnection;
            this.fClientSide = clientSide;
        }
        
        /**
         * constructor
         * @param aTServerSideProtocol the TTLSProvider created will be run
         *        on the server
         */
        public TTLSProvider(TServerSideProtocol aTServerSideProtocol) {
            this.fmyTTCPConnection = aTServerSideProtocol.conn();
            this.fClientSide = false;
        }
        
        /**
         * use this method to go back to the plain-text communications layer;
         * makes sense only if you have established an SSL session prior to this
         * method call
         * @throws java.io.IOException
         */
        public void makeInsecure() throws IOException {
            if (this.fmyTTCPConnection.mySocket() != null) {
                if (this.fmyTTCPConnection.mySocket() instanceof SSLSocket) {
                    ((SSLSocket)(this.fmyTTCPConnection.mySocket())).getSession().invalidate();
                    ((SSLSocket)(this.fmyTTCPConnection.mySocket())).close();
                    Socket aSocket = this.fmyTTCPConnection.myRegularSocket();
                    this.fmyTTCPConnection.connect(aSocket);                                                            
                }                
            }
        }


        public static SSLSocketFactory createSSLSocketFactory(KeyManager[] keyManagers, TrustManager[] trustManagers, SecureRandom secureRandom) {
            try {
                SSLContext sslContext = SSLContext.getInstance("TLSv1");

                if (trustManagers == null) {
                    trustManagers = new TrustManager[1];
                    trustManagers[0] = new TDummyCATrustManager();
                }
                if (trustManagers.length == 0) {
                    trustManagers = new TrustManager[1];
                    trustManagers[0] = new TDummyCATrustManager();
                }

                sslContext.init(keyManagers, trustManagers, secureRandom);

                SSLSocketFactory sslSocketFactory = ((SSLSocketFactory)(sslContext.getSocketFactory()));

                return sslSocketFactory;
            } catch (Exception e) {
                return null;
            }
        }

        /**
         * use this method to establish an SSL layer over an existing plain-text
         * connection; this method can be invoked on both the client side and
         * the server side, and it does not require the use of X509 certificates
         * @param protocols
         * @param ciphersuites
         * @param regularSocketAutoClose
         * @throws java.io.IOException
         * @throws java.security.NoSuchAlgorithmException
         * @throws java.security.KeyManagementException
         */
        public void makeSecureWithoutAuthentication(
                String[] protocols,
                String[] ciphersuites,
                boolean  regularSocketAutoClose                
                ) throws IOException, 
                         NoSuchAlgorithmException, 
                         KeyManagementException {                                           
            SSLSocketFactory anSSLSocketFactory = TTLSProvider.createSSLSocketFactory(null, null, null);
            this.makeSecureWithoutAuthentication(anSSLSocketFactory, protocols, ciphersuites, regularSocketAutoClose);
        }        

        public void makeSecureWithoutAuthentication(SSLSocketFactory anSSLSocketFactory) throws IOException, NoSuchAlgorithmException, KeyManagementException {
            this.makeSecureWithoutAuthentication(anSSLSocketFactory, null, null, false);
        }

        public void makeSecureWithoutAuthentication(
                SSLSocketFactory anSSLSocketFactory,
                String[] protocols,
                String[] ciphersuites,
                boolean  regularSocketAutoClose                
                ) throws IOException, 
                         NoSuchAlgorithmException, 
                         KeyManagementException {
            if (this.fmyTTCPConnection.mySocket() instanceof SSLSocket) {
                this.makeInsecure();
            }
            
            if (!(this.fmyTTCPConnection.isConnected())) {
                throw new IOException();
            }
            
            if (protocols == null) {
                protocols = new String[2];
                protocols[0] = "TLSv1";
                protocols[1] = "SSLv3";                        
            }
        
            if (ciphersuites == null) {
                ciphersuites = new String[3];
                ciphersuites[0] = "TLS_RSA_WITH_AES_256_CBC_SHA";
                ciphersuites[1] = "TLS_DH_anon_WITH_AES_256_CBC_SHA";          
                ciphersuites[2] = "SSL_DH_anon_WITH_3DES_EDE_CBC_SHA";                        
                
            }
                                            
            SSLSocket anSSLSocket = ((SSLSocket)(anSSLSocketFactory.createSocket(
                    this.fmyTTCPConnection.mySocket(),
                    this.fmyTTCPConnection.mySocket().getInetAddress().getHostName(),
                    this.fmyTTCPConnection.mySocket().getPort(),
                    regularSocketAutoClose)));                                                                 
         
            anSSLSocket.setEnabledProtocols(protocols);
            anSSLSocket.setEnabledCipherSuites(ciphersuites);
        
            anSSLSocket.setUseClientMode(this.fClientSide);                  
            if (!(this.fClientSide)) {
                anSSLSocket.setWantClientAuth(false);                
            }
            anSSLSocket.startHandshake();
        
            this.fmyTTCPConnection.connect(anSSLSocket);                        
        }

        /**
         * use this method to create an SSL tunnel in which at least the server
         * presents its X509 certificate to the client; client may also connect
         * with makeSecureWithoutAuthentication()
         * @param protocols
         * @param ciphersuites
         * @param keyManagers
         * @param trustManagers
         * @param regularSocketAutoClose
         * @throws java.io.IOException
         * @throws java.security.NoSuchAlgorithmException
         * @throws java.security.KeyManagementException
         */
        public void makeSecureWithServerCertAuthByClient(
                String[] protocols,
                String[] ciphersuites,
                KeyManager[] keyManagers,
                TrustManager[] trustManagers,                
                boolean  regularSocketAutoClose                
                ) throws IOException, 
                         NoSuchAlgorithmException, 
                         KeyManagementException {
            
            SSLSocketFactory anSSLSocketFactory = TTLSProvider.createSSLSocketFactory(keyManagers, trustManagers, null);
            this.makeSecureWithServerCertAuthByClient(anSSLSocketFactory, protocols, ciphersuites, regularSocketAutoClose);
        }        

        public void makeSecureWithServerCertAuthByClient(SSLSocketFactory anSSLSocketFactory) throws IOException, NoSuchAlgorithmException, KeyManagementException {
            this.makeSecureWithServerCertAuthByClient(anSSLSocketFactory, null, null, false);
        }

        public void makeSecureWithServerCertAuthByClient(
                SSLSocketFactory anSSLSocketFactory,
                String[] protocols,
                String[] ciphersuites,
                boolean  regularSocketAutoClose                
                ) throws IOException, 
                         NoSuchAlgorithmException, 
                         KeyManagementException {
            if (this.fmyTTCPConnection.mySocket() instanceof SSLSocket) {
                this.makeInsecure();
            }
            
            if (!(this.fmyTTCPConnection.isConnected())) {
                throw new IOException();
            }
            
            if (protocols == null) {
                protocols = new String[2];
                protocols[0] = "TLSv1";
                protocols[1] = "SSLv3";                        
            }
        
            if (ciphersuites == null) {
                ciphersuites = new String[3];
                ciphersuites[0] = "TLS_RSA_WITH_AES_256_CBC_SHA";
                ciphersuites[1] = "TLS_DH_anon_WITH_AES_256_CBC_SHA";          
                ciphersuites[2] = "SSL_DH_anon_WITH_3DES_EDE_CBC_SHA";                                        
            }
                                            
            SSLSocket anSSLSocket = ((SSLSocket)(anSSLSocketFactory.createSocket(
                    this.fmyTTCPConnection.mySocket(),
                    this.fmyTTCPConnection.mySocket().getInetAddress().getHostName(),
                    this.fmyTTCPConnection.mySocket().getPort(),
                    regularSocketAutoClose)));                                                                 
         
            anSSLSocket.setEnabledProtocols(protocols);
            anSSLSocket.setEnabledCipherSuites(ciphersuites);
        
            anSSLSocket.setUseClientMode(this.fClientSide);                  
            if (!(this.fClientSide)) {
                anSSLSocket.setWantClientAuth(false);                
            }
            anSSLSocket.startHandshake();
        
            this.fmyTTCPConnection.connect(anSSLSocket);                        
        }

        /**
         * use this method on the server to make sure that only clients 
         * connect that present their credentials (X509); no cert = no access.
         * @param protocols
         * @param ciphersuites
         * @param keyManagers
         * @param trustManagers
         * @param regularSocketAutoClose
         * @throws java.io.IOException
         * @throws java.security.NoSuchAlgorithmException
         * @throws java.security.KeyManagementException
         */
        public void makeSecureWithClientCertAuthByServer(
                String[] protocols,
                String[] ciphersuites,
                KeyManager[] keyManagers,
                TrustManager[] trustManagers,                
                boolean  regularSocketAutoClose                
                ) throws IOException, 
                         NoSuchAlgorithmException, 
                         KeyManagementException {                                
            SSLSocketFactory anSSLSocketFactory = TTLSProvider.createSSLSocketFactory(keyManagers, trustManagers, null);
            this.makeSecureWithClientCertAuthByServer(anSSLSocketFactory, protocols, ciphersuites, regularSocketAutoClose);
        }                

        public void makeSecureWithClientCertAuthByServer(SSLSocketFactory anSSLSocketFactory) throws IOException, NoSuchAlgorithmException, KeyManagementException {
            this.makeSecureWithClientCertAuthByServer(anSSLSocketFactory, null, null, false);
        }

        public void makeSecureWithClientCertAuthByServer(
                SSLSocketFactory anSSLSocketFactory,
                String[] protocols,
                String[] ciphersuites,
                boolean  regularSocketAutoClose
                ) throws IOException,
                         NoSuchAlgorithmException,
                         KeyManagementException {
            if (this.fmyTTCPConnection.mySocket() instanceof SSLSocket) {
                this.makeInsecure();
            }

            if (!(this.fmyTTCPConnection.isConnected())) {
                throw new IOException();
            }

            if (protocols == null) {
                protocols = new String[2];
                protocols[0] = "TLSv1";
                protocols[1] = "SSLv3";
            }

            if (ciphersuites == null) {
                ciphersuites = new String[1];
                ciphersuites[0] = "TLS_RSA_WITH_AES_256_CBC_SHA";
            }

            SSLSocket anSSLSocket = ((SSLSocket)(anSSLSocketFactory.createSocket(
                    this.fmyTTCPConnection.mySocket(),
                    this.fmyTTCPConnection.mySocket().getInetAddress().getHostName(),
                    this.fmyTTCPConnection.mySocket().getPort(),
                    regularSocketAutoClose)));

            anSSLSocket.setEnabledProtocols(protocols);
            anSSLSocket.setEnabledCipherSuites(ciphersuites);

            anSSLSocket.setUseClientMode(this.fClientSide);
            if (!(this.fClientSide)) {
                anSSLSocket.setNeedClientAuth(true);
            }
            anSSLSocket.startHandshake();

            this.fmyTTCPConnection.connect(anSSLSocket);
        }
}
